# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ptyp',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/ptyp',
                'lib_path': 'ptyp'},
  'syms': { 'ptyp.grds': { 'ptyp.grds.aliascheck': ('grds.html#aliascheck', 'ptyp/grds.py'),
                           'ptyp.grds.eqlist': ('grds.html#eqlist', 'ptyp/grds.py'),
                           'ptyp.grds.eqliteral': ('grds.html#eqliteral', 'ptyp/grds.py'),
                           'ptyp.grds.eqtuple': ('grds.html#eqtuple', 'ptyp/grds.py'),
                           'ptyp.grds.eqtype': ('grds.html#eqtype', 'ptyp/grds.py'),
                           'ptyp.grds.equnion': ('grds.html#equnion', 'ptyp/grds.py'),
                           'ptyp.grds.guard': ('grds.html#guard', 'ptyp/grds.py'),
                           'ptyp.grds.isallsame': ('grds.html#isallsame', 'ptyp/grds.py'),
                           'ptyp.grds.isstrbytes': ('grds.html#isstrbytes', 'ptyp/grds.py'),
                           'ptyp.grds.isstrq': ('grds.html#isstrq', 'ptyp/grds.py'),
                           'ptyp.grds.notiterstr': ('grds.html#notiterstr', 'ptyp/grds.py'),
                           'ptyp.grds.notstrbytes': ('grds.html#notstrbytes', 'ptyp/grds.py'),
                           'ptyp.grds.tclassvar': ('grds.html#tclassvar', 'ptyp/grds.py'),
                           'ptyp.grds.tdict': ('grds.html#tdict', 'ptyp/grds.py'),
                           'ptyp.grds.tgeneric': ('grds.html#tgeneric', 'ptyp/grds.py'),
                           'ptyp.grds.tlist': ('grds.html#tlist', 'ptyp/grds.py'),
                           'ptyp.grds.tliteral': ('grds.html#tliteral', 'ptyp/grds.py'),
                           'ptyp.grds.tset': ('grds.html#tset', 'ptyp/grds.py'),
                           'ptyp.grds.ttuple': ('grds.html#ttuple', 'ptyp/grds.py'),
                           'ptyp.grds.ttype': ('grds.html#ttype', 'ptyp/grds.py'),
                           'ptyp.grds.tunion': ('grds.html#tunion', 'ptyp/grds.py')},
            'ptyp.meta': { 'ptyp.meta.Alias': ('meta.html#alias', 'ptyp/meta.py'),
                           'ptyp.meta.AliasMeta': ('meta.html#aliasmeta', 'ptyp/meta.py'),
                           'ptyp.meta.AliasMeta.getbtype': ('meta.html#aliasmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.Not': ('meta.html#not', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta': ('meta.html#notmeta', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.check': ('meta.html#notmeta.check', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.default_kwds': ('meta.html#notmeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.getbtype': ('meta.html#notmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.Opt': ('meta.html#opt', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta': ('meta.html#optionalmeta', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.__new__': ('meta.html#optionalmeta.__new__', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.default_kwds': ('meta.html#optionalmeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.getbtype': ('meta.html#optionalmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtoType': ('meta.html#prototype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtoType.__init_subclass__': ('meta.html#prototype.__init_subclass__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta': ('meta.html#protocolbasemeta', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__call__': ('meta.html#protocolbasemeta.__call__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__instancecheck__': ('meta.html#protocolbasemeta.__instancecheck__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__new__': ('meta.html#protocolbasemeta.__new__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.check': ('meta.html#protocolbasemeta.check', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.classname': ('meta.html#protocolbasemeta.classname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.default_kwds': ('meta.html#protocolbasemeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getbtype': ('meta.html#protocolbasemeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.makebases': ('meta.html#protocolbasemeta.makebases', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.makeguard': ('meta.html#protocolbasemeta.makeguard', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.qualname': ('meta.html#protocolbasemeta.qualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setattrs': ('meta.html#protocolbasemeta.setattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setbtype': ('meta.html#protocolbasemeta.setbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.settypes': ('meta.html#protocolbasemeta.settypes', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolMeta': ('meta.html#protocolmeta', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolMeta.getbtype': ('meta.html#protocolmeta.getbtype', 'ptyp/meta.py')},
            'ptyp.prot': { 'ptyp.prot.Alias': ('prot.html#alias', 'ptyp/prot.py'),
                           'ptyp.prot.Alias.__init_subclass__': ('prot.html#alias.__init_subclass__', 'ptyp/prot.py'),
                           'ptyp.prot.Not': ('prot.html#not', 'ptyp/prot.py'),
                           'ptyp.prot.Not.__init_subclass__': ('prot.html#not.__init_subclass__', 'ptyp/prot.py'),
                           'ptyp.prot.Opt': ('prot.html#opt', 'ptyp/prot.py'),
                           'ptyp.prot.Opt.__init_subclass__': ('prot.html#opt.__init_subclass__', 'ptyp/prot.py'),
                           'ptyp.prot.Proto': ('prot.html#proto', 'ptyp/prot.py'),
                           'ptyp.prot.Proto.__init_subclass__': ('prot.html#proto.__init_subclass__', 'ptyp/prot.py')},
            'ptyp.ptyp': { 'ptyp.ptyp.IntQ': ('ptyp.html#intq', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.NotStrBytes': ('ptyp.html#notstrbytes', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.StrBytes': ('ptyp.html#strbytes', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.StrQ': ('ptyp.html#strq', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.isallsame': ('ptyp.html#isallsame', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.notiterstr': ('ptyp.html#notiterstr', 'ptyp/ptyp.py')},
            'ptyp.type': {},
            'ptyp.util': { 'ptyp.util.args2strs': ('util.html#args2strs', 'ptyp/util.py'),
                           'ptyp.util.classname': ('util.html#classname', 'ptyp/util.py'),
                           'ptyp.util.guardanns': ('util.html#guardanns', 'ptyp/util.py'),
                           'ptyp.util.guarddocs': ('util.html#guarddocs', 'ptyp/util.py'),
                           'ptyp.util.guardname': ('util.html#guardname', 'ptyp/util.py'),
                           'ptyp.util.guardtype': ('util.html#guardtype', 'ptyp/util.py'),
                           'ptyp.util.optargname': ('util.html#optargname', 'ptyp/util.py'),
                           'ptyp.util.qualname': ('util.html#qualname', 'ptyp/util.py'),
                           'ptyp.util.setguardattrs': ('util.html#setguardattrs', 'ptyp/util.py'),
                           'ptyp.util.settypes': ('util.html#settypes', 'ptyp/util.py'),
                           'ptyp.util.types2opts': ('util.html#types2opts', 'ptyp/util.py'),
                           'ptyp.util.typescsv': ('util.html#typescsv', 'ptyp/util.py'),
                           'ptyp.util.typesusv': ('util.html#typesusv', 'ptyp/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_meta.ipynb.

# %% auto 0
__all__ = ['ProtocolBaseMeta', 'ProtocolMeta', 'ProtoType', 'AliasMeta', 'Alias', 'NotMeta', 'Not', 'OptionalMeta', 'Opt']

# %% ../nbs/03_meta.ipynb 6
import os, inspect
from abc import abstractmethod
from functools import wraps, singledispatch

# %% ../nbs/03_meta.ipynb 8
from types import NoneType 
from typing import (Type, TypeGuard, Protocol, _ProtocolMeta, runtime_checkable, overload, )

# %% ../nbs/03_meta.ipynb 10
#| export


# %% ../nbs/03_meta.ipynb 12
#| export


# %% ../nbs/03_meta.ipynb 14
from .type import (T, Types, Guard, StrBool)
from .util import (classname, qualname, settypes)
from .grds import (aliascheck)

# %% ../nbs/03_meta.ipynb 17
class ProtocolBaseMeta(_ProtocolMeta):
    '''Base metaclass for creating protocol classes with custom type checking.

    This metaclass provides mechanisms to define custom type checks for classes 
    and to dynamically create classes with these checks.

    Attributes
    ----------
    types : Types
        A tuple of types that the class will check against.
    attrs : dict
        A dictionary of attributes that the class will check against.

    Methods
    -------
    getbtype()
        Abstract method to get the base type of the protocol.
    check(obj)
        Checks if an object conforms to the protocol.
    makeguard()
        Creates a guard method for the protocol.
    makebases(cls)
        Determines the base classes for a new class instance.
    default_kwds(**kwargs)
        Provides default keyword arguments for class creation.
    classname(cls, clsname)
        Gets the class name.
    qualname(cls, *types, clsname, prefix, suffix, dropnone)
        Generates a qualified name for the class.
    settypes(cls, *types, clsname, prefix, suffix, dropnone)
        Sets the types for the class.
    setattrs(cls, attrs)
        Sets the attributes for the class.
    setbtype(cls, btype)
        Sets the base type for the class.
    __call__(cls, *args, **kwargs)
        A callable method that enables the dynamic creation of classes.
    '''
    
    types: Types = tuple()
    attrs: dict = dict()

    @classmethod
    @abstractmethod
    def getbtype(cls):
        '''Abstract class method to get the base type of the protocol.
        
        Returns
        -------
        None
            Placeholder return type for the abstract method.
        '''
        return None
    
    def check(cls, obj) -> Guard:
        '''Class method to check if an object conforms to the protocol.

        Parameters
        ----------
        obj : Any
            The object to be checked against the protocol.

        Returns
        -------
        Guard
            The result of the type checking.
        '''
        return aliascheck(obj, cls.types, cls.attrs)    
    
    @classmethod
    def makeguard(mcls) -> Guard:
        '''Class method to create a guard method for the protocol.

        Returns
        -------
        Guard
            A class method acting as a guard for the protocol.
        '''
        return classmethod(lambda cls, obj: mcls.check(cls, obj))
        # return classmethod(lambda cls, obj: aliascheck(obj, cls.types, cls.attrs))
        return classmethod(lambda cls, obj: cls.check(obj))
    
    @classmethod
    def makebases(mcls, cls):
        '''Class method to determine the base classes for a new class instance.

        Parameters
        ----------
        cls : type
            The class for which the base classes are to be determined.

        Returns
        -------
        tuple
            A tuple of base classes.
        '''
        btype = mcls.getbtype()
        bases = (cls, btype) if btype not in cls.__bases__ else (cls,)
        return bases

    @classmethod
    def default_kwds(cls, **kwargs) -> dict:
        '''Class method to provide default keyword arguments for class creation.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments for class creation.

        Returns
        -------
        dict
            A dictionary of default keyword arguments.
        '''
        kwds = dict(
            clsname  = kwargs.get('clsname', True),
            prefix   = kwargs.get('prefix', None),
            suffix   = kwargs.get('suffix', None),
            dropnone = kwargs.get('dropnone', True),
        )
        return kwds

    def __new__(mcls, name, bases, dct, **kwargs):
        '''Class method to create a new class instance.

        Parameters
        ----------
        name : str
            The name of the new class.
        bases : tuple
            A tuple of base classes.
        dct : dict
            A dictionary of class attributes.
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        type
            A new class instance.
        '''
        kwds = mcls.default_kwds(**kwargs)
        
        dct['types'] = (types := kwargs.pop('types', getattr(mcls, 'types', tuple())))
        dct['attrs'] = (attrs := kwargs.pop('attrs', getattr(mcls, 'attrs', dict())))        
        # dct['btype'] = (btype := kwargs.pop('btype', getattr(mcls, 'btype', None)))
        # dct['check'] = classmethod(lambda cls, obj: aliascheck(obj, types, attrs))
        dct['check'] = mcls.makeguard()        
        new = super().__new__(mcls, name, bases, dct)
        new.__qualname__ = new.qualname(*types, **kwds)
        new.settypes(*types, **kwds)
        new.setattrs(attrs)
        # new.setbtype(attrs)
        
        return new
    
    def __instancecheck__(self, ins) -> Guard:
        '''Instance method to check if an instance conforms to the protocol.

        Parameters
        ----------
        ins : Any
            The instance to be checked.

        Returns
        -------
        Guard
            The result of the instance checking.
        '''
        return self.check(ins)
    
    def classname(cls, clsname: StrBool = True) -> StrBool:
        '''Class method to get the class name.

        Parameters
        ----------
        clsname : StrBool
            The class name or a boolean indicating if the class name should be used.

        Returns
        -------
        StrBool
            The class name or a boolean value.
        '''
        return classname(cls, clsname)
        
    def qualname(cls, *types: Types, clsname: StrBool = True, prefix: str = None, suffix: str = None, dropnone: bool = True) -> str:
        '''Class method to generate a qualified name for the class.

        Parameters
        ----------
        *types : Types
            A tuple of types.
        clsname : StrBool
            The class name or a boolean indicating if the class name should be used.
        prefix : str, optional
            A prefix for the qualified name.
        suffix : str, optional
            A suffix for the qualified name.
        dropnone : bool, optional
            A flag to indicate whether None types should be dropped.

        Returns
        -------
        str
            The qualified name of the class.
        '''
        return qualname(cls, *types, clsname=clsname, prefix=prefix, suffix=suffix, dropnone=dropnone)
    
    def settypes(cls, *types: Types, clsname: StrBool = True, prefix: str = None, suffix: str = None, dropnone: bool = True):
        '''Class method to set the types for the class.

        Parameters
        ----------
        *types : Types
            A tuple of types.
        clsname : StrBool
            The class name or a boolean indicating if the class name should be used.
        prefix : str, optional
            A prefix for the qualified name.
        suffix : str, optional
            A suffix for the qualified name.
        dropnone : bool, optional
            A flag to indicate whether None types should be dropped.
        '''
        settypes(cls, *types, clsname=clsname, prefix=prefix, suffix=suffix, dropnone=dropnone)
        
    def setattrs(cls, attrs: dict = dict()):
        '''Class method to set the attributes for the class.

        Parameters
        ----------
        attrs : dict, optional
            A dictionary of attributes to be set for the class.
        '''
        setattr(cls, 'attrs', attrs)

    def setbtype(cls, btype: Type = None):
        '''Class method to set the base type for the class.

        Parameters
        ----------
        btype : Type, optional
            The base type to be set for the class.
        '''
        setattr(cls, 'btype', btype)

    @overload
    def __call__(cls: 'ProtocolBaseMeta', base: type, *dtypes: Types, **kwargs) -> TypeGuard[T]: ...
    @overload
    def __call__(cls: type, instance: object) -> TypeGuard[T]: ...
    @overload
    def __call__(cls: type, *instances: tuple[object, ...]) -> TypeGuard[T]: ...
    @overload
    def __call__(cls: type, *args, **kwargs) -> TypeGuard[T]: ...    
    def __call__(cls: type, *args, **kwargs) -> TypeGuard[T]:
        '''Overloaded callable method that enables dynamic creation and type checking of classes.

        Parameters
        ----------
        *args : variable
            Variable arguments for different use cases.
        **kwargs : dict
            Keyword arguments for different use cases.

        Returns
        -------
        TypeGuard[T]
            Depending on the arguments, either performs type checking or creates a new class instance.
        '''
        if len(args) > 0 and not isinstance(args[0], type) or args[0] is NoneType:
            if len(args) == 1: return cls.check(args[0])
            if all(not inspect.isclass(i) for i in args):
                return tuple(cls.check(i) for i in args)
            return cls.check(args[0])
        
        dtypes = args
        def decorator(kls):
            '''Decorator to apply negation to a class.'''
            clsname = kwargs.get('clsname', kls.__name__)
            bases = cls.makebases(kls)
            dct = dict()
            new = cls.__class__(clsname, bases, dct, types=dtypes, **kwargs)
            return new
        return decorator

# %% ../nbs/03_meta.ipynb 20
class ProtocolMeta(ProtocolBaseMeta):
    '''Metaclass for creating protocol classes with a specific base type.

    Inherits from ProtocolBaseMeta and provides a specific base type for the 
    protocol classes created using this metaclass.

    Example
    -------
    >>> @Alias(str, bytes)
    ... class StrBytes: ...
    >>> StrBytes('hello')  # Checks if 'hello' is an instance of str or bytes.
    '''
    @classmethod
    def getbtype(cls): return ProtoType

@runtime_checkable
class ProtoType(Protocol):
    '''Base protocol class with custom type checking capabilities.

    This class provides mechanisms for custom type checking and can be used as a 
    base class for creating protocol types.

    Examples
    --------
    >>> @Not(str, bytes)
    ... class NotStrBytes: ...
    >>> NotStrBytes(123)  # Checks if 123 is not an instance of str or bytes.

    ### Subclassing Example
    >>> class MyType(ProtoType, types=(int, float)):
    ...     \'\'\'Custom type that checks against int and float\'\'\'
    >>> MyType(10)  # Checks if 10 is an instance of int or float.
    '''
    @classmethod
    def __init_subclass__(cls, *args, types: Types = tuple(), attrs: dict = dict(), **kwargs) -> None:
        super().__init_subclass__()
        kwds = cls.default_kwds(**attrs)
        cls.settypes(*types, **kwds)
        cls.setattrs(attrs)
        cls.setbtype(cls)

# %% ../nbs/03_meta.ipynb 22
class AliasMeta(ProtocolBaseMeta):
    '''Metaclass for creating alias classes with custom type checking.

    Inherits from ProtocolBaseMeta and provides a specific base type for the 
    alias classes created using this metaclass.

    Example
    -------
    >>> @Alias(str, bytes)
    ... class StrBytes: ...
    >>> StrBytes('example')  # Checks if 'example' is an instance of str or bytes.
    ''' 
    @classmethod
    def getbtype(cls): return Alias
    

@runtime_checkable
class Alias(ProtoType, Protocol, metaclass=AliasMeta):
    '''Class representing an alias with custom type checking.

    Inherits from ProtoType and Protocol, and is created using the AliasMeta 
    metaclass. It represents an alias with specific type checking behavior.

    Example
    -------
    >>> @Alias(str, bytes)
    ... class StrBytes: ...
    >>> StrBytes(b'bytes')  # Checks if b'bytes' is an instance of str or bytes.
    '''
    ...

# %% ../nbs/03_meta.ipynb 24
class NotMeta(ProtocolBaseMeta):
    '''Metaclass for creating negation classes with custom type checking.

    Inherits from ProtocolBaseMeta and overrides certain methods to provide 
    negation logic in type checking for the classes created using this metaclass.

    Example
    -------
    >>> @Not(str, bytes)
    ... class NotStrBytes: ...
    >>> NotStrBytes(['not', 'str', 'or', 'bytes'])  # Checks if list is not an instance of str or bytes.
    '''
    @classmethod
    def getbtype(cls): return Not

    @classmethod
    def default_kwds(cls, **kwargs) -> dict:
        kwds = super().default_kwds(**kwargs)
        kwds['prefix'] = '~'        
        return kwds
    
    def check(self, obj) -> Guard:     
        return not aliascheck(obj, self.types, self.attrs)
    
@runtime_checkable
class Not(ProtoType, Protocol, metaclass=NotMeta):
    '''Class representing a negation with custom type checking.

    Inherits from ProtoType and Protocol, and is created using the NotMeta 
    metaclass. It represents a negation type with specific type checking behavior.

    Example
    -------
    >>> @Not(str, bytes)
    ... class NotStrBytes: ...
    >>> NotStrBytes(100)  # Checks if 100 is not an instance of str or bytes.
    '''
    ...

# %% ../nbs/03_meta.ipynb 26
class OptionalMeta(ProtocolBaseMeta):
    '''Metaclass for creating optional type classes with custom type checking.

    Inherits from ProtocolBaseMeta and modifies the creation process to include 
    NoneType in the types for the classes created using this metaclass.

    Example
    -------
    >>> @Opt(str)
    ... class StrQ: ...
    >>> StrQ(None)  # Checks if None is an instance of str or NoneType.
    >>> StrQ('optional')  # Checks if 'optional' is an instance of str or NoneType.
    '''
    @classmethod
    def getbtype(cls): return Opt

    @classmethod
    def default_kwds(cls, **kwargs) -> dict:
        kwds = super().default_kwds(**kwargs)
        kwds['prefix'] = '?'        
        return kwds
    
    def __new__(mcls, name, bases, dct, **kwargs):
        '''Create a new class instance.'''
        types = kwargs.pop('types', tuple())
        for ntype in (None, ):
            if ntype not in types:
                types += (ntype, )
        dct['types'] = types        
        new = super().__new__(mcls, name, bases, dct, types = types, **kwargs)
        return new
    
@runtime_checkable
class Opt(ProtoType, Protocol, metaclass=OptionalMeta):
    '''Class representing an optional type with custom type checking.

    Inherits from ProtoType and Protocol, and is created using the OptionalMeta 
    metaclass. It represents an optional type with specific type checking behavior.

    Example
    -------
    >>> @Opt(int)
    ... class IntQ: ...
    >>> IntQ(None)  # Checks if None is an instance of int or NoneType.
    >>> IntQ(42)  # Checks if 42 is an instance of int or NoneType.
    '''
    ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_type.ipynb.

# %% auto 0
__all__ = ['T', 'P', 'Types', 'TypeLike', 'Guard', 'GuardFunc', 'StrFactory', 'StrBool']

# %% ../nbs/00_type.ipynb 6
import os, inspect
from functools import wraps

# %% ../nbs/00_type.ipynb 8
from typing import (Tuple, Union, Callable, Type, TypeVar, TypeAlias, TypeGuard,  ParamSpec,)

# %% ../nbs/00_type.ipynb 10
#| export


# %% ../nbs/00_type.ipynb 12
#| export


# %% ../nbs/00_type.ipynb 14
#| export


# %% ../nbs/00_type.ipynb 16
T = TypeVar('T')
P = ParamSpec('P')

Types: TypeAlias = Tuple[Type, ...]
'''`*args` type''';

TypeLike: TypeAlias = Union[type, Type, Type[T], TypeAlias]
'''type, Type, type of type, TypeAlias, or `*args` type''';

Guard: TypeAlias = TypeGuard[TypeLike]
'''Generic TypeGuard''';

GuardFunc: TypeAlias = Callable[P, Guard]
'''Generic TypeGuard function''';

StrFactory: TypeAlias = Callable[[Types], str]
'''Generic string factory that takes any number of types and returns a string describing it''';

StrBool: TypeAlias = Union[str, bool]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_prot.ipynb.

# %% auto 0
__all__ = ['Proto', 'Alias', 'Not', 'Opt']

# %% ../nbs/04_prot.ipynb 6
import os, inspect
from functools import wraps

# %% ../nbs/04_prot.ipynb 8
from typing import (
    Any, Set, Dict, List, Self, Tuple, Union, ClassVar, Generic,
    Literal, _LiteralGenericAlias, 
    Callable, Iterable, Optional, 
    Protocol, _ProtocolMeta,
    get_args, get_origin, runtime_checkable, 
    Type, TypeVar, TypeAlias, TypeGuard,  ParamSpec,
)


from types import NoneType 

# %% ../nbs/04_prot.ipynb 10
#| export


# %% ../nbs/04_prot.ipynb 12
#| export


# %% ../nbs/04_prot.ipynb 14
from .type import (T, P, Types, TypeLike, Guard, GuardFunc, StrFactory, StrBool)
from ptyp.util import (
    args2strs, typescsv, typesusv,
    guardtype,
    guardname, guarddocs, guardanns, setguardattrs,
    classname, qualname, settypes
)

from ptyp.grds import (
    guard,
    tset, tlist, tdict, ttype, ttuple, tunion, tliteral, tgeneric, tclassvar,
    eqtype, eqlist, eqtuple, equnion, eqliteral,
    isstrbytes, notstrbytes, isstrq, notiterstr, allsame,
    aliascheck
)

from ptyp.meta import (
    ProtocolBaseMeta, ProtocolMeta, AliasMeta, NotMeta, OptionalMeta
)

# %% ../nbs/04_prot.ipynb 16
@runtime_checkable
class Proto(Protocol, metaclass=ProtocolMeta):
    @classmethod
    def __init_subclass__(cls, *args, types: Types = tuple(), attrs: dict = dict(), **kwargs) -> None:
        super().__init_subclass__()
        kwds = cls.default_kwds(**attrs)
        cls.settypes(*types, **kwds)
        cls.setattrs(attrs)

@runtime_checkable
class Alias(Protocol, metaclass=AliasMeta):
    @classmethod
    def __init_subclass__(cls, *args, types: Types = tuple(), attrs: dict = dict(), **kwargs) -> None:
        super().__init_subclass__()
        kwds = cls.default_kwds(**attrs)
        cls.settypes(*types, **kwds)
        cls.setattrs(attrs)

@runtime_checkable
class Not(Protocol, metaclass=NotMeta):
    @classmethod
    def __init_subclass__(cls, *args, types: Types = tuple(), attrs: dict = dict(), **kwargs) -> None:
        super().__init_subclass__()
        kwds = cls.default_kwds(**attrs)
        cls.settypes(*types, **kwds)
        cls.setattrs(attrs)


@runtime_checkable
class Opt(Protocol, metaclass=OptionalMeta):
    @classmethod
    def __init_subclass__(cls, *args, types: Types = tuple(), attrs: dict = dict(), **kwargs) -> None:
        super().__init_subclass__()
        kwds = cls.default_kwds(**attrs)
        cls.settypes(*types, **kwds)
        cls.setattrs(attrs)
